# -*- coding: utf-8 -*-
"""Supernova oscillation physics for flavors e, X, e-bar, X-bar.

For measured mixing angles and latest global analysis results, visit
http://www.nu-fit.org/.
"""

from abc import abstractmethod, ABC

import numpy as np
from astropy import units as u
from astropy import constants as c

from .neutrino import MassHierarchy, MixingParameters


class FlavorTransformation(ABC):
    """Generic interface to compute neutrino and antineutrino survival probability."""

    @abstractmethod
    def prob_ee(self, t, E):
        """Electron neutrino survival probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        float or ndarray
            Transition probability.
        """
        pass

    @abstractmethod
    def prob_ex(self, t, E):
        """X -> e neutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        float or ndarray
            Transition probability.
        """
        pass

    @abstractmethod
    def prob_xx(self, t, E):
        """Flavor X neutrino survival probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        float or ndarray
            Transition probability.
        """
        pass

    @abstractmethod
    def prob_xe(self, t, E):
        """e -> X neutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        float or ndarray
            Transition probability.
        """
        pass

    @abstractmethod
    def prob_eebar(self, t, E):
        """Electron antineutrino survival probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        float or ndarray
            Transition probability.
        """
        pass

    @abstractmethod
    def prob_exbar(self, t, E):
        """X -> e antineutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        float or ndarray
            Transition probability.
        """
        pass

    @abstractmethod
    def prob_xxbar(self, t, E):
        """X -> X antineutrino survival probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        float or ndarray
            Transition probability.
        """
        pass

    @abstractmethod
    def prob_xebar(self, t, E):
        """e -> X antineutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        float or ndarray
            Transition probability.
        """
        pass


class NoTransformation(FlavorTransformation):
    """Survival probabilities for no oscillation case."""

    def __init__(self):
        pass

    def prob_ee(self, t, E):
        """Electron neutrino survival probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return 1.    

    def prob_ex(self, t, E):
        """X -> e neutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return 1. - self.prob_ee(t,E)

    def prob_xx(self, t, E):
        """Flavor X neutrino survival probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return (1. + self.prob_ee(t,E)) / 2.   

    def prob_xe(self, t, E):
        """e -> X neutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return (1. - self.prob_ee(t,E)) / 2.

    def prob_eebar(self, t, E):
        """Electron antineutrino survival probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return 1.

    def prob_exbar(self, t, E):
        """X -> e antineutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return 1. - self.prob_eebar(t,E)       

    def prob_xxbar(self, t, E):
        """X -> X antineutrino survival probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return (1. + self.prob_eebar(t,E)) / 2.

    def prob_xebar(self, t, E):
        """e -> X antineutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return (1. - self.prob_eebar(t,E)) / 2.

    
class CompleteExchange(FlavorTransformation):
    """Survival probabilities for the case when the electron flavors are completely exchanged with the x flavor."""

    def __init__(self):
        pass

    def prob_ee(self, t, E):
        """Electron neutrino survival probability.
        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.
        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return 0.    

    def prob_ex(self, t, E):
        """X -> e neutrino transition probability.
        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.
        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return 1. - self.prob_ee(t,E)

    def prob_xx(self, t, E):
        """Flavor X neutrino survival probability.
        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.
        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return (1. + self.prob_ee(t,E)) / 2.   

    def prob_xe(self, t, E):
        """e -> X neutrino transition probability.
        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.
        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return (1. - self.prob_ee(t,E)) / 2.

    def prob_eebar(self, t, E):
        """Electron antineutrino survival probability.
        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.
        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return 0.

    def prob_exbar(self, t, E):
        """X -> e antineutrino transition probability.
        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.
        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return 1. - self.prob_eebar(t,E)       

    def prob_xxbar(self, t, E):
        """X -> X antineutrino survival probability.
        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.
        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return (1. + self.prob_eebar(t,E)) / 2.

    def prob_xebar(self, t, E):
        """e -> X antineutrino transition probability.
        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.
        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return (1. - self.prob_eebar(t,E)) / 2.


class AdiabaticMSW(FlavorTransformation):
    """Adiabatic MSW effect."""

    def __init__(self, mix_angles=None, mh=MassHierarchy.NORMAL):
        """Initialize transformation matrix.

        Parameters
        ----------
        mix_angles : tuple or None
            If not None, override default mixing angles using tuple (theta12, theta13, theta23).
        mh : MassHierarchy
            MassHierarchy.NORMAL or MassHierarchy.INVERTED.
        """
        if type(mh) == MassHierarchy:
            self.mass_order = mh
        else:
            raise TypeError('mh must be of type MassHierarchy')

        if mix_angles is not None:
            theta12, theta13, theta23 = mix_angles
        else:
            pars = MixingParameters(mh)
            theta12, theta13, theta23 = pars.get_mixing_angles()

        self.De1 = float((np.cos(theta12) * np.cos(theta13))**2)
        self.De2 = float((np.sin(theta12) * np.cos(theta13))**2)
        self.De3 = float(np.sin(theta13)**2)

    def prob_ee(self, t, E):
        """Electron neutrino survival probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        if self.mass_order == MassHierarchy.NORMAL:
            return self.De3
        else:
            return self.De2

    def prob_ex(self, t, E):
        """X -> e neutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return 1. - self.prob_ee(t,E)

    def prob_xx(self, t, E):
        """Flavor X neutrino survival probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return (1. + self.prob_ee(t,E)) / 2.   

    def prob_xe(self, t, E):
        """e -> X neutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return (1. - self.prob_ee(t,E)) / 2.

    def prob_eebar(self, t, E):
        """Electron antineutrino survival probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        if self.mass_order == MassHierarchy.NORMAL:
            return self.De1  
        else:
            return self.De3

    def prob_exbar(self, t, E):
        """X -> e antineutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return 1. - self.prob_eebar(t,E)       

    def prob_xxbar(self, t, E):
        """X -> X antineutrino survival probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return (1. + self.prob_eebar(t,E)) / 2.

    def prob_xebar(self, t, E):
        """e -> X antineutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return (1. - self.prob_eebar(t,E)) / 2.


class NonAdiabaticMSWH(FlavorTransformation):
    """Nonadiabatic MSW effect."""

    def __init__(self, mix_angles=None, mh=MassHierarchy.NORMAL):
        """Initialize transformation matrix.

        Parameters
        ----------
        mix_angles : tuple or None
            If not None, override default mixing angles using tuple (theta12, theta13, theta23).
        mh : MassHierarchy
            MassHierarchy.NORMAL or MassHierarchy.INVERTED.
        """
        if type(mh) == MassHierarchy:
            self.mass_order = mh
        else:
            raise TypeError('mh must be of type MassHierarchy')

        if mix_angles is not None:
            theta12, theta13, theta23 = mix_angles
        else:
            pars = MixingParameters(mh)
            theta12, theta13, theta23 = pars.get_mixing_angles()

        self.De1 = float((np.cos(theta12) * np.cos(theta13))**2)
        self.De2 = float((np.sin(theta12) * np.cos(theta13))**2)
        self.De3 = float(np.sin(theta13)**2)

    def prob_ee(self, t, E):
        """Electron neutrino survival probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return self.De2

    def prob_ex(self, t, E):
        """X -> e neutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return 1. - self.prob_ee(t,E)

    def prob_xx(self, t, E):
        """Flavor X neutrino survival probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return (1. + self.prob_ee(t,E)) / 2.   

    def prob_xe(self, t, E):
        """e -> X neutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return (1. - self.prob_ee(t,E)) / 2.

    def prob_eebar(self, t, E):
        """Electron antineutrino survival probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return self.De1  

    def prob_exbar(self, t, E):
        """X -> e antineutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return 1. - self.prob_eebar(t,E)       

    def prob_xxbar(self, t, E):
        """X -> X antineutrino survival probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return (1. + self.prob_eebar(t,E)) / 2.

    def prob_xebar(self, t, E):
        """e -> X antineutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return (1. - self.prob_eebar(t,E)) / 2.


class TwoFlavorDecoherence(FlavorTransformation):
    """Star-earth transit survival probability: two flavor case."""

    def __init__(self, mix_angles=None, mh=MassHierarchy.NORMAL):
        """Initialize transformation matrix.

        Parameters
        ----------
        mix_angles : tuple or None
            If not None, override default mixing angles using tuple (theta12, theta13, theta23).
        mh : MassHierarchy
            MassHierarchy.NORMAL or MassHierarchy.INVERTED.
        """
        if type(mh) == MassHierarchy:
            self.mass_order = mh
        else:
            raise TypeError('mh must be of type MassHierarchy')

        if mix_angles is not None:
            theta12, theta13, theta23 = mix_angles
        else:
            pars = MixingParameters(mh)
            theta12, theta13, theta23 = pars.get_mixing_angles()

        self.De1 = float((np.cos(theta12) * np.cos(theta13))**2)
        self.De2 = float((np.sin(theta12) * np.cos(theta13))**2)
        self.De3 = float(np.sin(theta13)**2)

    def prob_ee(self, t, E):
        """Electron neutrino survival probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.
        """
        if self.mass_order == MassHierarchy.NORMAL:
            return (self.De2+self.De3)/2.    
        else:
            return self.De2

    def prob_ex(self, t, E):
        """X -> e neutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return 1. - self.prob_ee(t,E)

    def prob_xx(self, t, E):
        """Flavor X neutrino survival probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return (1. + self.prob_ee(t,E)) / 2.   

    def prob_xe(self, t, E):
        """e -> X neutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return (1. - self.prob_ee(t,E)) / 2.

    def prob_eebar(self, t, E):
        """Electron antineutrino survival probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        if self.mass_order == MassHierarchy.NORMAL:
            return self.De1
        else:
            return (self.De1+self.De3)/2

    def prob_exbar(self, t, E):
        """X -> e antineutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return 1. - self.prob_eebar(t,E)       

    def prob_xxbar(self, t, E):
        """X -> X antineutrino survival probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return (1. + self.prob_eebar(t,E)) / 2.

    def prob_xebar(self, t, E):
        """e -> X antineutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return (1. - self.prob_eebar(t,E)) / 2.


class ThreeFlavorDecoherence(FlavorTransformation):
    """Star-earth transit survival probability: three flavor case."""

    def __init__(self):
        pass

    def prob_ee(self, t, E):
        """Electron neutrino survival probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.
        """
        return 1./3.    

    def prob_ex(self, t, E):
        """X -> e neutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return 1. - self.prob_ee(t,E)

    def prob_xx(self, t, E):
        """Flavor X neutrino survival probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return (1. + self.prob_ee(t,E)) / 2.   

    def prob_xe(self, t, E):
        """e -> X neutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return (1. - self.prob_ee(t,E)) / 2.  

    def prob_eebar(self, t, E):
        """Electron antineutrino survival probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return 1./3.

    def prob_exbar(self, t, E):
        """X -> e antineutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return 1. - self.prob_eebar(t,E)       

    def prob_xxbar(self, t, E):
        """X -> X antineutrino survival probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return (1. + self.prob_eebar(t,E)) / 2.

    def prob_xebar(self, t, E):
        """e -> X antineutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return (1. - self.prob_eebar(t,E)) / 2.


class NeutrinoDecay(FlavorTransformation):
    """Decay effect, where the heaviest neutrino decays to the lightest
    neutrino. For a description and typical parameters, see A. de Gouvêa et al.,
    PRD 101:043013, 2020, arXiv:1910.01127.
    """
    def __init__(self, mix_angles=None, mass=1*u.eV/c.c**2, tau=1*u.day, dist=10*u.kpc, mh=MassHierarchy.NORMAL):
        """Initialize transformation matrix.

        Parameters
        ----------
        mix_angles : tuple or None
            If not None, override default mixing angles using tuple (theta12, theta13, theta23).
        mass : astropy.units.quantity.Quantity
            Mass of the heaviest neutrino; expect in eV/c^2.
        tau : astropy.units.quantity.Quantity
            Lifetime of the heaviest neutrino.
        dist : astropy.units.quantity.Quantity
            Distance to the supernova.
        mh : MassHierarchy
            MassHierarchy.NORMAL or MassHierarchy.INVERTED.
        """
        if type(mh) == MassHierarchy:
            self.mass_order = mh
        else:
            raise TypeError('mh must be of type MassHierarchy')

        if mix_angles is not None:
            theta12, theta13, theta23 = mix_angles
        else:
            pars = MixingParameters(mh)
            theta12, theta13, theta23 = pars.get_mixing_angles()

        self.De1 = float((np.cos(theta12) * np.cos(theta13))**2)
        self.De2 = float((np.sin(theta12) * np.cos(theta13))**2)
        self.De3 = float(np.sin(theta13)**2)

        self.m = mass
        self.tau = tau
        self.d = dist

    def gamma(self, E):
        """Decay width of the heaviest neutrino mass.

        Parameters
        ----------
        E : float
            Energy of the nu3.

        Returns
        -------
        Gamma : float
            Decay width of the neutrino mass, in units of 1/length.

        :meta private:
        """
        return self.m*c.c / (E*self.tau)

    def prob_ee(self, t, E):
        """Electron neutrino survival probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        # NMO case.
        if self.mass_order == MassHierarchy.NORMAL:
            pe_array = self.De1*(1-np.exp(-self.gamma(E)*self.d)) + \
                       self.De3*np.exp(-self.gamma(E)*self.d)
        # IMO case.
        else:
            pe_array = self.De2*np.exp(-self.gamma(E)*self.d) + \
                       self.De3*(1-np.exp(-self.gamma(E)*self.d))
        return pe_array

    def prob_ex(self, t, E):
        """X -> e neutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        # NMO case.
        if self.mass_order == MassHierarchy.NORMAL:
            return self.De1 + self.De3
        # IMO case.
        else:
            return self.De1 + self.De2

    def prob_xx(self, t, E):
        """Flavor X neutrino survival probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return 1. - self.prob_ex(t,E) / 2.

    def prob_xe(self, t, E):
        """e -> X neutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return (1. - self.prob_ee(t,E)) / 2.

    def prob_eebar(self, t, E):
        """Electron antineutrino survival probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return self.De3  

    def prob_exbar(self, t, E):
        """X -> e antineutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        # NMO case.
        if self.mass_order == MassHierarchy.NORMAL:
            pxbar_array = self.De1*(1-np.exp(-self.gamma(E)*self.d)) + \
                          self.De2 + self.De3*np.exp(-self.gamma(E)*self.d)
        # IMO case.
        else:
            pxbar_array = self.De1 + self.De2*np.exp(-self.gamma(E)*self.d) + \
                          self.De3*(1-np.exp(-self.gamma(E)*self.d))
        return pxbar_array

    def prob_xxbar(self, t, E):
        """X -> X antineutrino survival probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return 1. - self.prob_exbar(t,E) / 2.

    def prob_xebar(self, t, E):  
        """e -> X antineutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return (1. - self.prob_eebar(t,E)) / 2.       


class AdiabaticMSWes(FlavorTransformation):
    """A four-neutrino mixing prescription. The assumptions used are that:

    1. the fourth neutrino mass is the heaviest but not so large that the electron-sterile resonances
       are inside the neutrinosphere;
    2. the “outer” or H' electron-sterile MSW resonance is adiabatic;
    3. the “inner” or H'' electron-sterile MSW resonance (where the electron fraction = 1/3) is non-adiabatic.

    For further insight see, for example, Esmaili, Peres, and Serpico, Phys. Rev. D 90, 033013 (2014).
    """
    def __init__(self, mix_angles, mh=MassHierarchy.NORMAL):
        """Initialize transformation matrix.

        Parameters
        ----------
        mix_angles : tuple
            Values for mixing angles (theta12, theta13, theta23, theta14).
        mh : MassHierarchy
            MassHierarchy.NORMAL or MassHierarchy.INVERTED.
        """
        if type(mh) == MassHierarchy:
            self.mass_order = mh
        else:
            raise TypeError('mh must be of type MassHierarchy')

        theta12, theta13, theta23, theta14 = mix_angles

        self.De1 = float((np.cos(theta12) * np.cos(theta13) * np.cos(theta14))**2)
        self.De2 = float((np.sin(theta12) * np.cos(theta13) * np.cos(theta14))**2)
        self.De3 = float((np.sin(theta13) * np.cos(theta14))**2)
        self.De4 = float((np.sin(theta14))**2)
        self.Ds1 = float((np.cos(theta12) * np.cos(theta13) * np.sin(theta14))**2)
        self.Ds2 = float((np.sin(theta12) * np.cos(theta13) * np.sin(theta14))**2)
        self.Ds3 = float((np.sin(theta13) * np.sin(theta14))**2)
        self.Ds4 = float((np.cos(theta14))**2)
    
    def prob_ee(self, t, E):
        """e -> e neutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return self.De4       
        
    def prob_ex(self, t, E):
        """x -> e neutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """       
        if self.mass_order == MassHierarchy.NORMAL:
            return self.De1 + self.De2
        else:
            return self.De1 + self.De3     
        
    def prob_xx(self, t, E):
        """x -> x neutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """        
        if self.mass_order == MassHierarchy.NORMAL:
            return ( 2 - self.De1 - self.De2 - self.Ds1 - self.Ds2 ) / 2 
        else:
            return ( 2 - self.De1 - self.De3 - self.Ds1 - self.Ds3 ) / 2          
        
    def prob_xe(self, t, E):
        """e -> x neutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """        
        return ( 1 - self.De4 - self.Ds4 )/2         
    
    def prob_eebar(self, t, E):
        """e -> e antineutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """        
        if self.mass_order == MassHierarchy.NORMAL:
            return self.De1
        else:
            return self.De3
        
    def prob_exbar(self, t, E):
        """x -> e antineutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """        
        if self.mass_order == MassHierarchy.NORMAL:
            return self.De3 + self.De4
        else:
            return self.De2 + self.De4   
        
    def prob_xxbar(self, t, E):
        """x -> x antineutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """        
        if self.mass_order == MassHierarchy.NORMAL:
            return ( 2 - self.De3 - self.De4 - self.Ds3 - self.Ds4 ) / 2 
        else:
            return ( 2 - self.De2 - self.De4 - self.Ds2 - self.Ds4 ) / 2     
        
    def prob_xebar(self, t, E):
        """e -> x antineutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """        
        if self.mass_order == MassHierarchy.NORMAL:
            return ( 1 - self.De1 - self.Ds1 ) / 2
        else:
            return ( 1 - self.De3 - self.Ds3 ) / 2     


class NonAdiabaticMSWes(FlavorTransformation):
    """A four-neutrino mixing prescription. The assumptions used are that:

    1. the fourth neutrino mass is the heaviest but not so large that the electron-sterile resonances
       are inside the neutrinosphere;
    2. the “outer” or H' electron-sterile MSW resonance is non-adiabatic;
    3. the “inner” or H'' electron-sterile MSW resonance (where the electron fraction = 1/3) is non-adiabatic.

    For further insight see, for example, Esmaili, Peres, and Serpico, Phys. Rev. D 90, 033013 (2014).
    """
    def __init__(self, mix_angles, mh=MassHierarchy.NORMAL):
        """Initialize transformation matrix.
        
        Parameters
        ----------
        mix_angles : tuple
            Values for mixing angles (theta12, theta13, theta23, theta14).
        mh : MassHierarchy
            MassHierarchy.NORMAL or MassHierarchy.INVERTED.
        """
        if type(mh) == MassHierarchy:
            self.mass_order = mh
        else:
            raise TypeError('mh must be of type MassHierarchy')

        theta12, theta13, theta23, theta14 = mix_angles

        self.De1 = float((np.cos(theta12) * np.cos(theta13) * np.cos(theta14))**2)
        self.De2 = float((np.sin(theta12) * np.cos(theta13) * np.cos(theta14))**2)
        self.De3 = float((np.sin(theta13) * np.cos(theta14))**2)
        self.De4 = float((np.sin(theta14))**2)
        self.Ds1 = float((np.cos(theta12) * np.cos(theta13) * np.sin(theta14))**2)
        self.Ds2 = float((np.sin(theta12) * np.cos(theta13) * np.sin(theta14))**2)
        self.Ds3 = float((np.sin(theta13) * np.sin(theta14))**2)
        self.Ds4 = float((np.cos(theta14))**2)
    
    def prob_ee(self, t, E):
        """e -> e neutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """                
        if self.mass_order == MassHierarchy.NORMAL:
            return self.De3
        else:
            return self.De2        
        
    def prob_ex(self, t, E):
        """x -> e neutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """                
        if self.mass_order == MassHierarchy.NORMAL:
            return self.De1 + self.De2
        else:
            return self.De1 + self.De3        
        
    def prob_xx(self, t, E):
        """x -> x neutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """               
        if self.mass_order == MassHierarchy.NORMAL:
            return ( 2 - self.De1 - self.De2 - self.Ds1 - self.Ds2 ) / 2 
        else:
            return ( 2 - self.De1 - self.De3 - self.Ds1 - self.Ds3 ) / 2         
        
    def prob_xe(self, t, E):
        """e -> x neutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """                
        if self.mass_order == MassHierarchy.NORMAL:
            return (1 - self.De3 - self.Ds3)/2
        else:
            return (1 - self.De2 - self.Ds2) / 2        
    
    def prob_eebar(self, t, E):
        """e -> e antineutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """               
        if self.mass_order == MassHierarchy.NORMAL:
            return self.De1
        else:
            return self.De3    
        
    def prob_exbar(self, t, E):
        """x -> e antineutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """        
        if self.mass_order == MassHierarchy.NORMAL:
            return self.De2 + self.De3
        else:
            return self.De1 + self.De2        
        
    def prob_xxbar(self, t, E):
        """x -> x antineutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """        
        if self.mass_order == MassHierarchy.NORMAL:
            return ( 2 - self.De2 - self.De3 - self.Ds2 - self.Ds3 ) / 2 
        else:
            return ( 2 - self.De1 - self.De2 - self.Ds1 - self.Ds2 ) / 2        
        
    def prob_xebar(self, t, E):
        """e -> x antineutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """        
        if self.mass_order == MassHierarchy.NORMAL:
            return ( 1 - self.De1 - self.Ds1 ) / 2
        else:
            return ( 1 - self.De3 - self.Ds3 ) / 2        
    

class QuantumDecoherence(FlavorTransformation):
    """Quantum Decoherence, where propagation in vacuum leads to equipartition
    of states. For a description and typical parameters, see M. V. dos Santos et al.,
    2023, arXiv:2306.17591.
    """
    def __init__(self, mix_angles=None, Gamma3=1e-27*u.eV, Gamma8=1e-27*u.eV, dist=10*u.kpc, n=0, E0=10*u.MeV, mh=MassHierarchy.NORMAL):
        """Initialize transformation matrix.

        Parameters
        ----------
        mix_angles : tuple or None
            If not None, override default mixing angles using tuple (theta12, theta13, theta23).
        Gamma3 : astropy.units.quantity.Quantity
            Quantum decoherence parameter; expect in eV.
        Gamma8 : astropy.units.quantity.Quantity
            Quantum decoherence parameter; expect in eV.
        dist : astropy.units.quantity.Quantity
            Distance to the supernova.
        n : float
            Exponent of power law for energy dependent quantum decoherence parameters,
            i.e. Gamma = Gamma0*(E/E0)**n. If not specified, it is taken as zero.
        E0 : astropy.units.quantity.Quantity
            Reference energy in the power law Gamma = Gamma0*(E/E0)**n. If not specified, 
            it is taken as 10 MeV. Note that if n = 0, quantum decoherence parameters are independent
            of E0.
        mh : MassHierarchy
            MassHierarchy.NORMAL or MassHierarchy.INVERTED.
        """
        if type(mh) == MassHierarchy:
            self.mass_order = mh
        else:
            raise TypeError('mh must be of type MassHierarchy')

        if mix_angles is not None:
            theta12, theta13, theta23 = mix_angles
        else:
            pars = MixingParameters(mh)
            theta12, theta13, theta23 = pars.get_mixing_angles()

        self.De1 = float((np.cos(theta12) * np.cos(theta13))**2)
        self.De2 = float((np.sin(theta12) * np.cos(theta13))**2)
        self.De3 = float(np.sin(theta13)**2)

        self.Gamma3 = (Gamma3 / (c.hbar.to('eV s') * c.c)).to('1/kpc')
        self.Gamma8 = (Gamma8 / (c.hbar.to('eV s') * c.c)).to('1/kpc')
        self.d = dist
        self.n = n
        self.E0 = E0
    
    def P11(self, E):
        """Survival probability of state nu1 in vacuum.

        Parameters
        ----------
        E : float
            Energy.

        Returns
        -------
        P11 : float
            Survival probability of state nu1 in vacuum.

        :meta private:
        """
        return 1/3 + 1/2 * np.exp(-(self.Gamma3 * (E/self.E0)**self.n + self.Gamma8 * (E/self.E0)**self.n / 3) * self.d) + 1/6 * np.exp(-self.Gamma8 * (E/self.E0)**self.n * self.d)

    def P21(self, E):
        """Transition probability from the state nu2 to nu1 in vacuum.

        Parameters
        ----------
        E : float
            Energy.

        Returns
        -------
        P21 : float
            Transition probability from the state nu2 to nu1 in vacuum.
            Note that P21 = P12.

        :meta private:
        """
        return 1/3 - 1/2 * np.exp(-(self.Gamma3 * (E/self.E0)**self.n + self.Gamma8 * (E/self.E0)**self.n / 3) * self.d) + 1/6 * np.exp(-self.Gamma8 * (E/self.E0)**self.n * self.d)

    def P22(self, E):
        """Survival probability of state nu2 in vacuum.

        Parameters
        ----------
        E : float
            Energy.

        Returns
        -------
        P21 : float
            Survival probability of state nu2 in vacuum.

        :meta private:
        """
        return self.P11(E)


    def P31(self, E):
        """Transition probability from the state nu3 to nu1 in vacuum.

        Parameters
        ----------
        E : float
            Energy.

        Returns
        -------
        P31 : float
            Transition probability from the state nu3 to nu1 in vacuum.
            Note that P31 = P13.

        :meta private:
        """
        return 1/3 - 1/3 * np.exp(-self.Gamma8 * (E/self.E0)**self.n * self.d)
    
    def P32(self, E):
        """Transition probability from the state nu3 to nu2 in vacuum.

        Parameters
        ----------
        E : float
            Energy.

        Returns
        -------
        P32 : float
            Transition probability from the state nu3 to nu2 in vacuum.
            Note that P32 = P23.

        :meta private:
        """
        return self.P31(E)
    
    def P33(self, E):
        """Survival probability of state nu3 in vacuum.

        Parameters
        ----------
        E : float
            Energy.

        Returns
        -------
        P33 : float
            Survival probability of state nu3 in vacuum.

        :meta private:
        """
        return 1/3 + 2/3 * np.exp(-self.Gamma8 * (E/self.E0)**self.n * self.d)

    def prob_ee(self, t, E):
        """Electron neutrino survival probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        # NMO case.
        if self.mass_order == MassHierarchy.NORMAL:
            pe_array = self.P31(E)*self.De1 + self.P32(E)*self.De2 + self.P33(E)*self.De3
        # IMO case.
        else:
            pe_array = self.P22(E)*self.De2 + self.P21(E)*self.De1 + self.P32(E)*self.De3
        return pe_array

    def prob_ex(self, t, E):
        """X -> e neutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return 1. - self.prob_ee(t,E)

    def prob_xx(self, t, E):
        """Flavor X neutrino survival probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return 1. - self.prob_ex(t,E) / 2.

    def prob_xe(self, t, E):
        """e -> X neutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return (1. - self.prob_ee(t,E)) / 2.

    def prob_eebar(self, t, E):
        """Electron antineutrino survival probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        # NMO case.
        if self.mass_order == MassHierarchy.NORMAL:
            pe_array = self.P11(E)*self.De1 + self.P21(E)*self.De2 + self.P31(E)*self.De3
        # IMO case.
        else:
            pe_array = self.P31(E)*self.De1 + self.P32(E)*self.De2 + self.P33(E)*self.De3
        return pe_array

    def prob_exbar(self, t, E):
        """X -> e antineutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return 1. - self.prob_eebar(t,E)

    def prob_xxbar(self, t, E):
        """X -> X antineutrino survival probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return 1. - self.prob_exbar(t,E) / 2.

    def prob_xebar(self, t, E):  
        """e -> X antineutrino transition probability.

        Parameters
        ----------
        t : float or ndarray
            List of times.
        E : float or ndarray
            List of energies.

        Returns
        -------
        prob : float or ndarray
            Transition probability.
        """
        return (1. - self.prob_eebar(t,E)) / 2.



# -*- coding: utf-8 -*-
"""
The submodule ``snewpy.models.presn`` contains models of presupernova neutrino fluxes,
derived from the :class:`SupernovaModel` base class.
"""
import snewpy.models.presn_loaders as loaders
from snewpy.models.registry_model import RegistryModel
from astropy import units as u

@RegistryModel(
    progenitor_mass = [15, 25]<<u.Msun
)
class Odrzywolek_2010(loaders.Odrzywolek_2010):
    """Presupernova model based on
    `A.Odrzywolek, Alexander Heger, Acta Phys.Polon.B 41 (2010) <https://inspirehep.net/literature/870759>`_
    
    Dataset available on `Odrzywolek’s website <http://th.if.uj.edu.pl/~odrzywolek/psns/index.html>`_
    """
    def __init__(self, progenitor_mass:u.Quantity):
        filename=f"s{progenitor_mass.to_value('Msun'):.0f}_nuebar_data.txt"
        super().__init__(filename)
    
@RegistryModel(
    progenitor_mass = [15, 30]<<u.Msun
)
class Patton_2017(loaders.Patton_2017):
    """Presupernova model based on
    `Kelly M. Patton et al. 2017 ApJ 851 6 <https://iopscience.iop.org/article/10.3847/1538-4357/aa95c4>`_
    
    Dataset available on Zenodo (`DOI:10.5281/zenodo.2626645 <http://doi.org/10.5281/zenodo.2626645>`_)
    """
    def __init__(self, progenitor_mass:u.Quantity):
        filename=f"totalLuminosity_{progenitor_mass.to_value('Msun'):.0f}SolarMass.dat"
        super().__init__(filename)
        
@RegistryModel(
    progenitor_mass = [12, 15]<<u.Msun
)
class Kato_2017(loaders.Kato_2017):
    """Presupernova model based on
    `Chinami Kato et al. 2017 ApJ 848 48 <https://iopscience.iop.org/article/10.3847/1538-4357/aa8b72>`_
    
    Dataset available on `Zenodo <https://zenodo.org/records/3768052>`__
    """
    def __init__(self, progenitor_mass:u.Quantity):
        path=f"pre_collapse/m{progenitor_mass.to_value('Msun'):.0f}"
        super().__init__(path)


@RegistryModel(
    progenitor_mass = [12, 15, 20]<<u.Msun
)
class Yoshida_2016(loaders.Yoshida_2016):
    """Presupernova model based on
    `Yoshida et al. (2016), PRD 93, 123012 <https://doi.org/10.1103/PhysRevD.93.123012>`_
    
    Dataset available on `Zenodo <https://zenodo.org/records/3778014>`__
    """
    def __init__(self, progenitor_mass:u.Quantity):
        path=f"t_spc_m{progenitor_mass.to_value('Msun'):.0f}_1.2.txt"
        super().__init__(path)



"""
The module :mod:`snewpy.flux` defines the class :class:`snewpy.flux.Container` - a container for the neutrino flux, fluence, event rates etc.

This object wraps a 3D array, and its dimensions: `flavor`, `time` and `energy`.

Usage
-----
The flux container will be produced by the :meth:`SupernovaModel.get_flux`, and consumed by the :meth:`RateCalculator.run`

A example of usage:

.. code-block:: python

    import astropy.units as u
    import numpy as np
    from snewpy.models import ccsn
    from snewpy.flavor_transformation import AdiabaticMSW
    
    #get the suernova model
    model = ccsn.Bollig_2016(progenitor_mass=27<<u.Msun)
    #define the sampling points
    energy = np.linspace(0,100,51)<<u.MeV
    time = np.linspace(0,10,1000)<<u.s
    
    #calculate the flux
    flux = model.get_flux(time, energy, distance=10<<u.kpc, flavor_xform=AdiabaticMSW())

    #optionally: integrate the flux over time bins to obtain the fluence
    fluence = flux.integrate('time', limits=np.linspace(0,1,21)<<u.s)
    
    #calculate the event rates using the RateCalculator
    from snewpy.rate_calculator import RateCalculator
    rc = RateCalculator()
    event_rates = rc.run(fluence, detector='icecube')
    ibd_rate = event_rates['ibd'] #will be also a Container instance
    #get the total values vs. time bin
    N_ibd_vs_T = ibd_rate.sum('energy')
    #or total values vs. energy bin
    N_ibd_vs_E = ibd_rate.sum('time')
    #or total number of interactions
    N_ibd_tot = ibd_rate.sum('time').sum('energy')
    #to retrieve the number we can access the `array` member
    print(N_ibd_tot.array.squeeze()) #312249.3525844854

Reference
---------

.. autoclass:: Container
    :inherited-members:

.. autoclass:: Axes

"""
from typing import Union, Optional, Set, List
from snewpy.neutrino import Flavor
from astropy import units as u

import numpy as np

from scipy.integrate import cumulative_trapezoid
from scipy.interpolate import interp1d
from enum import IntEnum
from functools import wraps

#list of units which will be used as units for decomposition inside the Container
snewpy_unit_bases = [u.MeV, u.m, u.s, u.kg]

class Axes(IntEnum):
    """Enum to keep the number number of the array dimension for each axis""" 
    flavor=0, #Flavor dimension
    time=1,   #Time dimension
    energy=2, #Energy dimension
    
    @classmethod
    def get(cls, value:Union['Axes',int,str])->'Axes':
        "convert string,int or Axes value to Axes"
        if isinstance(value,str):
            return cls[value]
        else:
            return cls(value)

class _ContainerBase:
    """base class for internal use
    :noindex:
    """
    unit = None
    def __init__(self, 
                 data: u.Quantity,
                 flavor: List[Flavor],
                 time: u.Quantity[u.s], 
                 energy: u.Quantity[u.MeV],
                 *,
                 integrable_axes: Optional[Set[Axes]] = None
    ):
        """A container class storing the physical quantity (flux, fluence, rate...), which depends on flavor, time and energy.

        Parameters
        ----------
        data: :class:`astropy.Quantity`
            3D array of the stored quantity, must have dimensions compatible with (flavor, time, energy)
        
        flavor: list or a single value of :class:`snewpy.neutrino.Flavor`
            array of flavors (should be ``len(flavor)==data.shape[0]``
        
        time: :class:`astropy.Quantity`
            sampling points in time (then ``len(time)==data.shape[1]``) 
            or time bin edges (then ``len(time)==data.shape[1]+1``) 
    
        energy: :class:`astropy.Quantity`
            sampling points in energy (then ``len(energy)=data.shape[2]``) 
            or energy bin edges (then ``len(energy)=data.shape[2]+1``) 
    
        integrable_axes: set of :class:`Axes` or None
            List of axes which can be integrated.
            If None (default) this set will be derived from the axes shapes 
        """
        if self.unit is not None:
            #try to convert to the unit
            data = data.to(self.unit)
        #convert the input values to arrays if they are scalar
        self.array = u.Quantity(data)
        self.time = u.Quantity(time, ndmin=1)
        self.energy = u.Quantity(energy, ndmin=1)
        self.flavor = np.sort(np.array(flavor, ndmin=1))
        
        Nf,Nt,Ne = len(self.flavor), len(self.time), len(self.energy)
        #list all valid shapes of the input array
        expected_shapes=[(nf,nt,ne) for nf in (Nf,Nf-1) for nt in (Nt,Nt-1) for ne in (Ne,Ne-1)]
        #treat special case if data is 1d array
        if self.array.ndim==1:
            #try to reshape the array to expected shape
            for expected_shape in expected_shapes:
                if np.prod(expected_shape)==self.array.size:
                    self.array = self.array.reshape(expected_shape)
                    break
        #validate the data array shape
        if self.array.shape not in expected_shapes:
            raise ValueError(f"Data array of shape {data.shape} is inconsistent with any valid shapes {expected_shapes}")
        
        if integrable_axes is not None:
            #store which axes can be integrated
            self._integrable_axes = set(integrable_axes)
        else:
            #guess which axes can be integrated
            self._integrable_axes = {a for a in Axes if(self._axshape[a]==self.array.shape[a])}
            self._integrable_axes.discard(Axes.flavor)
            
    @property
    def _sumable_axes(self):
        return set(Axes).difference(self._integrable_axes)
    @property
    def axes(self):
        return self.flavor, self.time, self.energy

    @property
    def _axshape(self):
        return tuple(len(a) for a in self.axes)

    @property
    def shape(self):
        return self.array.shape
        
    def __getitem__(self, args)->'Container':
        """Slice the flux array and produce a new Flux object"""
        try: 
            iter(args)
        except TypeError:
            args = [args]
        args = [a if isinstance(a, slice) else slice(a, a + 1) for a in args]
        #expand args to match axes
        args+=[slice(None)]*(len(Axes)-len(args))
        array = self.array.__getitem__(tuple(args))
        newaxes = [ax.__getitem__(arg) for arg, ax in zip(args, self.axes)]
        return self.__class__(array, *newaxes)

    def __repr__(self) -> str:
        """print information about the container"""
        s = [f"{len(values)} {label.name}({values.min()};{values.max()})"
            for label, values in zip(Axes,self.axes)]
        return f"{self.__class__.__name__} {self.array.shape} [{self.array.unit}]: <{' x '.join(s)}>"
    
    def sum(self, axis: Union[Axes,str])->'Container':
        """Sum along given axis, producing a Container with the summary quantity.
        
        Parameters
        -----------
            axis: :class:`Axes` or str
                An axis to sum over. String should be one of ``"flavor"``, ``"time"`` or ``"energy"``  (check :meth:`Container.can_sum`)
        Returns
        --------
            Container with summed value
                
        Raises
        ------
            ValueError 
                if the given axis cannot be summed over (i.e. it must be integrated instead, see :meth:`Container.integrate`).
                One can check summable axes with :attr:`Container._sumable_axes`

        Example
        -------
        The resulting data array will be 3D array, but the dimension, corresponding to `axis` parameter will be reduced to 1. 
        
        For an examplar Container ``a`` of a given shape::
        
            >>> a.shape
            (4, 10, 20)
            >>> a.sum('flavor').shape
            (1, 10, 20)
        
        The axis in the class will also be modified, keeping only the first and last points of the summation::
        
            >>> a.flavor
            array([0, 1, 2, 3])
            >>> a.sum('flavor').flavor
            array([0, 3])
            
        All the other axes and dimensions will be kept the same
        """
        axis = Axes.get(axis)
        if axis not in self._sumable_axes:
            raise ValueError(f'Cannot sum over {axis.name}! Valid axes are {self._sumable_axes}')
        array = np.sum(self.array, axis = axis, keepdims=True)
        axes = list(self.axes)
        axes[axis] = axes[axis].take([0,-1])
        return Container(array,*axes, integrable_axes = self._integrable_axes.difference({axis}))

    def integrate(self, axis:Union[Axes,str], limits:np.ndarray=None)->'Container':
        """Integrate along given axis, producing a Container with the integral quantity.
        
        Parameters
        -----------
            axis: :class:`Axes` or str
                An axis to integrate over. String should be one of ``"time"`` or ``"energy"`` (check :meth:`Container.can_integrate`)

            limits: np.ndarray or None
                A sorted array (or `astropy.Quantity` consistent with the units of given axis) of integration limits
                If limits are None (default), then integrate over the the whole range of this axis
                Otherwise limits are treated as bin edges - the integration  happens within each bin limits
                
        Returns
        --------
            Container with integrated value

        Raises
        ------
            ValueError 
                if the given axis cannot be integrated over (i.e. it must be summed instead, see :meth:`Container.sum`).
        Example
        -------
        The resulting data array will be 3D array, but the dimension, corresponding to `axis` parameter will be reduced to ``len(limits)-1``. 
        
        For an examplar Container ``a`` of a given shape::
        
            >>> a.shape
            (4, 10, 20)
            >>> a.integrate('time').shape
            (4, 1, 20)
            >>> a.integrate('time', limits=[0, 0.5, 1]<<u.s).shape
            (4, 2, 20)

        The axis in the class will also be modified, keeping only the integration limits::
        
            >>> a.time
            [0. 1. 2. 3. 4. 5. 6. 7. 8. 9.] s
            >>> a.integrate('time').time
            [0., 9.] s
            >>> a.integrate('time', limits=[0, 0.5, 1]<<u.s).time
            [0., 0.5, 1.] s
        
        All the other axes and dimensions will be kept the same
        """
        axis = Axes.get(axis)
        if not axis in self._integrable_axes:
            raise ValueError(f'Cannot integrate over {axis.name}! Valid axes are {self._integrable_axes}')
        #set the limits
        ax = self.axes[axis]
        xmin, xmax = ax.min(), ax.max()
        if limits is None:
            limits = u.Quantity([xmin, xmax])
        limits = limits.to(ax.unit)
        #limits = limits.clip(xmin,xmax)
        #compute the integral
        yc = cumulative_trapezoid(self.array, x=ax, axis=axis, initial=0)
        #get first and last value to use as the fill values
        yc_limits = (yc.take(0,axis=axis), yc.take(-1,axis=axis)) 
        #this will make the _integral constant if it gets out of bounds,
        # i.e. effectively the flux outside of bounds is zero
        _integral = interp1d(x=ax, y=yc, fill_value=yc_limits, axis=axis, bounds_error=False)
        array = np.diff(_integral(limits),axis=axis) << (self.array.unit*ax.unit)
        axes = list(self.axes)
        axes[axis] = limits
        #choose the proper class
        return Container(array, *axes, integrable_axes=self._integrable_axes.difference({axis}))

    def integrate_or_sum(self, axis:Union[Axes,str])->'Container':
        if self.can_integrate(axis):
            return self.integrate(axis)
        else:
            return self.sum(axis)
            
    def can_integrate(self, axis):
        "return true if can be integrated along given axis"
        return Axes.get(axis) in self._integrable_axes
    def can_sum(self, axis):
        "return true if can be summed along given axis"
        return Axes.get(axis) not in self._integrable_axes
    
    def __rmul__(self, factor):
        "multiply array by given factor or matrix"
        return self.__mul__(factor)

    def __mul__(self, factor) -> 'Container':
        "multiply array by given factor or matrix"
        if not (np.isscalar(factor) or isinstance(factor, np.ndarray)):
            return NotImplemented
        #    raise ValueError("Factor should be a scalar value")
        array = self.array*factor
        axes = list(self.axes)
        return Container(array, *axes)

    def save(self, fname:str)->None:
        """Save container data to a given file (using `numpy.savez`)"""
        def _save_quantity(name):
            values = self.__dict__[name]
            try:
                array = values.to_value()
                unit = values.unit.to_string()
                return {name:array, f'_{name}_unit':unit}
            except:
                return {name:values}
        data_dict = {}
        for name in ['array','time','energy','flavor']:
            data_dict.update(_save_quantity(name))
        np.savez(fname,
                 _class_name=self.__class__.__name__, 
                 **data_dict,
                 _integrable_axes=np.array([int(a) for a in self._integrable_axes])
                )
    
    @classmethod
    def load(cls, fname:str)->'Container':
        """Load container from a given file"""
        with np.load(fname) as f:
            def _load_quantity(name):
                array = f[name]
                try:
                    unit = str(f[f'_{name}_unit'])
                    return array<<u.Unit(unit)
                except KeyError:
                    return array
                    
            class_name = str(f['_class_name'])
            array = _load_quantity('array')
            cls = Container[array.unit, class_name]
            return cls(data=array,
                       **{name:_load_quantity(name) for name in ['time','energy','flavor']},
                       integrable_axes=f['_integrable_axes'])
            
    def __eq__(self, other:'Container')->bool:
        "Check if two Containers are equal"
        result = self.__class__==other.__class__ and \
                 self.unit == other.unit and \
                 np.allclose(self.array, other.array) and \
                 all([np.allclose(self.axes[ax], other.axes[ax]) for ax in Axes])
        return result

class Container(_ContainerBase):
    #a dictionary holding classes for each unit
    _unit_classes = {}

    @wraps(_ContainerBase.__init__)
    def __new__(cls, data,*args, **kwargs):
        data = data.decompose(snewpy_unit_bases) #simplify the units, reducing to the bases
        if not cls.unit:
            data = u.Quantity(data)
            cls = cls[data.unit]
        return super().__new__(cls)

    def __class_getitem__(cls, args):
        try:
            unit,name = args
        except:
            unit,name = args, None
        #convert units to astropy.Unit
        unit = u.Unit(unit)
        if unit not in cls._unit_classes:
            #create subclass of this type with given unit
            name = name or f'{cls.__name__}[{unit}]'
            cls._unit_classes[unit] = type(name,(cls,),{'unit':unit})
        return cls._unit_classes[unit]
        

#some standard container classes that can be used for 
Flux = Container['1/(MeV*s*m**2)', "d2FdEdT"]
Fluence = Container[Flux.unit*u.s, "dFdE"]
Spectrum= Container[Flux.unit*u.MeV, "dFdT"]
IntegralFlux= Container[Flux.unit*u.s*u.MeV, "dF"]

DifferentialEventRate = Container['1/(MeV*s)', "d2NdEdT"]
EventRate = Container['1/s', "dNdT"]
EventSpectrum = Container['1/MeV', "dNdE"]
EventNumber = Container['', "N"]



from warnings import warn

import numpy as np
from astropy import units as u

from snewpy.neutrino import Flavor
from snewpy.models.base import SupernovaModel


class ExtendedModel(SupernovaModel):
    """Class defining a supernova model with a cooling tail extension."""

    def __init__(self, base_model):
        """Initialize extended supernova model class."""
        if not isinstance(base_model, SupernovaModel):
            raise TypeError("ExtendedModel.__init__ requires a SupernovaModel object")

        self.__dict__ = base_model.__dict__.copy()
        for method_name in dir(base_model):
            if callable(getattr(base_model, method_name)) and method_name[0] != '_':
                if method_name == 'get_initial_spectra':
                    self._get_initial_spectra = getattr(base_model, method_name)
                else:
                    setattr(self, method_name, getattr(base_model, method_name))
        self.t_final = self.time[-1]
        self.L_final = {flv: self.luminosity[flv][-1] for flv in Flavor}

    def get_initial_spectra(self, *args, **kwargs):
        """Get neutrino spectra/luminosity curves before oscillation"""
        return self._get_initial_spectra(*args, **kwargs)

    def get_extended_luminosity(self, t, k=-1., A=None, tau_c=36. * u.s, alpha=2.66, flavor = Flavor.NU_E):
        """Get neutrino luminosity from supernova cooling tail luminosity model.

        Parameters
        ----------
        t : astropy.Quantity
            Time to evaluate luminosity.
        k : float
            Power law factor (default: -1)
        A : astropy.Quantity
            Normalization factor (default: None, automatically match original model data)
        tau_c : astropy.Quantity
            Exponential decay characteristic timescale (default: 36 s)
        alpha : float
            Exponential decay factor (default: 2.66)

        Returns
        -------
        astropy.Quantity
            Luminosity calculated from cooling tail model.
        """
        if t.value < 0.5:
            warn("Extended luminosity model not applicable to early times")
        if A is None:
            tf = self.t_final
            Lf = self.L_final[flavor]
            A = Lf / (tf.value**k * np.exp(-(tf/tau_c)**alpha))
        return A * t.value**k * np.exp(-(t/tau_c)**alpha)

    def extend(self, ts, k=-1., A=None, tau_c=36. * u.s, alpha=2.66):
        """Extend supernova model to specific times.

        Parameters
        ----------
        ts : astropy.Quantity
            Times to add to supernova model.
        k : float
            Power law factor (default: -1)
        A : astropy.Quantity
            Normalization factor (default: None, automatically match original model data)
        tau_c : astropy.Quantity
            Exponential decay characteristic timescale (default: 36 s)
        alpha : float
            Exponential decay factor (default: 2.66)
        """
        # Select times after the end of the model
        select = ts > self.t_final

        for t in ts[select]:
            self.time = np.append(self.time, t)
            for flavor in Flavor:
                L_ext = self.get_extended_luminosity(t, k = k, A = A, tau_c = tau_c, alpha = alpha, flavor = flavor)
                self.luminosity[flavor] = np.append(self.luminosity[flavor], L_ext)
                self.meanE[flavor] = np.append(self.meanE[flavor], self.meanE[flavor][-1])
                self.pinch[flavor] = np.append(self.pinch[flavor], self.pinch[flavor][-1])



import os
from warnings import warn

from snewpy.snowglobes import simulate

def go(SNOwGLoBESdir, Models_Path, Tarball, detector_input="all", verbose=False):
    # Deprecated since SNEWPY v1.1
    warn(f"The 'run_snowglobes.go()' function is deprecated. Use 'snowglobes.simulate()' instead.", FutureWarning)
    tarball_path = os.path.join(Models_Path, Tarball)
    return simulate(SNOwGLoBESdir, tarball_path, detector_input=detector_input)



# -*- coding: utf-8 -*-
"""Unit tests for the neutrino submodule.
"""

import unittest

from snewpy.neutrino import Flavor, MassHierarchy, MixingParameters

from astropy import units as u

class TestNeutrino(unittest.TestCase):

    def test_flavor(self):
        """
        Neutrino flavor types
        """
        nue  = Flavor.NU_E
        self.assertTrue(nue.is_electron)
        self.assertTrue(nue.is_neutrino)
        self.assertFalse(nue.is_antineutrino)

        nux  = Flavor.NU_X
        self.assertFalse(nux.is_electron)
        self.assertTrue(nux.is_neutrino)
        self.assertFalse(nux.is_antineutrino)

        nueb = Flavor.NU_E_BAR
        self.assertTrue(nueb.is_electron)
        self.assertFalse(nueb.is_neutrino)
        self.assertTrue(nueb.is_antineutrino)

        nuxb = Flavor.NU_X_BAR
        self.assertFalse(nuxb.is_electron)
        self.assertFalse(nuxb.is_neutrino)
        self.assertTrue(nuxb.is_antineutrino)


    def test_mixing_nmo(self):
        """
        Mixing parameter values; NMO
        """
        # By default, return mixing parameters for NMO.
        mixpars = MixingParameters(version="NuFIT5.0")
        self.assertEqual(mixpars.theta12, 33.44 * u.deg)
        self.assertEqual(mixpars.theta13,  8.57 * u.deg)
        self.assertEqual(mixpars.theta23, 49.20 * u.deg)
        self.assertEqual(mixpars.deltaCP, 197 * u.deg)
        self.assertEqual(mixpars.dm21_2, 7.42e-5 * u.eV**2)
        self.assertEqual(mixpars.dm31_2, 2.517e-3 * u.eV**2)


    def test_mixing_imo(self):
        """
        Mixing parameter values; IMO
        """
        mixpars = MixingParameters(MassHierarchy.INVERTED, version="NuFIT5.0")
        self.assertEqual(mixpars.theta12, 33.45 * u.deg)
        self.assertEqual(mixpars.theta13,  8.60 * u.deg)
        self.assertEqual(mixpars.theta23, 49.30 * u.deg)
        self.assertEqual(mixpars.deltaCP, 282 * u.deg)
        self.assertEqual(mixpars.dm21_2, 7.42e-5 * u.eV**2)
        self.assertEqual(mixpars.dm32_2, -2.498e-3 * u.eV**2)



"""
The module :mod:`snewpy.rate_calculator` defines a Python interface for calculating event rates using data from SNOwGLoBES

Reference
---------
.. autoclass:: RateCalculator
    :members: run
"""
import numpy as np
from snewpy.snowglobes_interface import SnowglobesData, guess_material
from snewpy.neutrino import Flavor
from snewpy.flux import Container
from astropy import units as u
from warnings import warn
from typing import Dict
from typing import Callable
import scipy.stats as st

#various utility methods
def center(a):
    return 0.5*(a[1:]+a[:-1])

u.kt = u.def_unit('kt',represents=1e6<<u.kg, doc='kilotonne')

class SmearingMatrix:
    """A class describing energy smearing, i.e. matrix transforming flux from :math:`E_\nu` (true neutrino energy) to :math:`E_{det}` (observed or smeared energy) space""" 
    def __init__(self, bins_true:u.Quantity, bins_smeared:u.Quantity, matrix:np.ndarray):
        """
        Parameters
        ----------
        bins_true: Quantity[energy]
            array of true energy bin limits
        bins_smeared: Quantity[energy]
            array of smeared (detected) energy bin limits
        matrix: 2D numpy array
            Smearing matrix of shape ``[len(bins_true)-1,len(bins_smeared)-1]``
        """
        self.bins_true = bins_true
        self.bins_smeared = bins_smeared
        self.matrix = matrix
        assert matrix.shape==(len(bins_true)-1,len(bins_smeared)-1), \
        f"Matrix shape {matrix.shape} is inconsistent with (bins_true-1, bins_smeared-1)={len(bins_true)-1,len(bins_smeared)-1}"

    def apply(self, rate:Container)-> Container:
        """Apply the smearing to the given rate/flux object, 
        and produce a smeared one

        Parameters
        ----------
        rate: Container
            Input rate/flux object in true energy space.
            If the rate is integrated over energy (binned), the energy bins must be equal to :attr:`self.bins_true`.

        Returns
        -------
        Container
            The rate/flux in the smeared energy space. 
            The energy bins of the returned container are equal to :attr:`self.bins_smeared`
        """
        if rate.can_integrate('energy'):
            rate = rate.integrate('energy', self.bins_true)
        assert np.allclose(rate.energy<<u.GeV, self.bins_true<<u.GeV), \
        f"rate.energy: {rate.energy<<u.GeV}!={self.bins_true<<u.GeV}"
        rateS_array = np.dot(rate.array, self.matrix)
        rateS = Container(rateS_array,
                          flavor=rate.flavor, 
                          time=rate.time, 
                          energy=self.bins_smeared,
                          integrable_axes=rate._integrable_axes)
        return rateS

    @classmethod
    def from_Gaussian(cls, bins_true:u.Quantity, bins_smeared:u.Quantity,
                      mean:Callable, sigma:Callable=1<<u.keV):
        """Construct a smearing matrix corresponding to the Gaussian smearing, with given mean and sigma.
        
        Parameters
        ----------
        bins_true: Quantity[energy]
            array of true energy bin limits
        bins_smeared: Quantity[energy]
            array of smeared (detected) energy bin limits
        mean: Callable or u.Quantity[energy]
            Mean value of the smeared energy. 
            Can be a function of E_nu, an array (of size `len(bins_true)-1`) or a scalar.
        sigma: Callable or u.Quantity[energy]
            Sigma of the smeared energy distribution.
            Can be a function of E_nu, an array (of size `len(bins_true)-1`) or a scalar.

        Examples
        --------
        
        >>> e_true =np.linspace(0,30,501)<<u.MeV
        >>> e_smear=np.linspace(0,30,501)<<u.MeV
        >>> #IBD process with minimal smearing, but with constant shift E_det = E_nu-0.87 MeV
        >>> smear_IBD = SmearingMatrix.from_Gaussian(e_true,e_smear, mean=lambda e_nu:e_nu-0.87*u.MeV)

        >>> #NC on C12 process with large smearing and constant energy response at 15.1 MeV
        >>> smear_NC_C = SmearingMatrix.from_Gaussian(e_true,e_smear, mean=15.1*u.MeV, sigma=2*u.MeV)

        """
        e_t = center(bins_true)<<u.MeV
        e_s = bins_smeared<<u.MeV
        
        if callable(mean):
            mean = mean(e_t)
        if callable(sigma):
            sigma = sigma(center(bins_true))
        #expand the scalar to required size
        mean = np.ones(e_t.shape)*mean
        sigma = np.ones(e_t.shape)*sigma
        #true energy will be axis=0    
        mean = np.atleast_2d(mean).T<<u.MeV
        sigma = np.atleast_2d(sigma).T<<u.MeV
        #smeared energy will be axis=1
        e_s = np.atleast_2d(e_s)
        distr =  st.norm(loc=mean, scale=sigma)
        #calculate integral in each bin
        cdf = distr.cdf(e_s)
        pdf = np.diff(cdf, axis=1)
        return cls(bins_true, bins_smeared, pdf)

class FunctionOfEnergy:
    """A wrapper around a given function of energy, for example interaction cross-section, or a detection efficiency.
    This helper class will perform a correct operation when multiplied to a :class:`Container` object - rate or flux.
    """
    def __init__(self, callable):
        """
        Parameters
        ----------
        callable: Callable
            A function of one parameter (energy). This can be an analitical function, or an interpolation of a (E,value) dataset
        """
        self.value = callable
    def __mul__(self, f:Container)->Container:
        e = f.energy #Define sample points
        if not f.can_integrate('energy'): #we have bins, let's use central values for sampling
            e = center(f.energy)
        return f*self.value(e)
    def __rmul__(self, f:Container)->Container:
        #same as multiplication from the left
        return self.__mul__(f)
        
    def __call__(self, energy):
        return self.value(energy)
        
    @classmethod
    def from_threshold(cls, e_min=1<<u.MeV):
        return cls(lambda e: 1*(e>e_min))

class DetectionChannel:
    """Description of a single detection channel in the detector"""
    
    def __init__(self, flavor:Flavor, xsec:callable, smearing:SmearingMatrix=None,
                 efficiency:FunctionOfEnergy=1., weight:float=1.):
        """
        Parameters
        ----------
        name:str
            channel name
        flavor:Flavor
            flavor of the interacting neutrino
        xsec:callable or FunctionOfEnergy
            crossection as a function of energy
        smearing:SmearingMatrix
            The detection energy smearing matrix. 
            If `None`(default) no smearing is applied
        efficiency: float or np.ndarray or callable.
            The detection efficiency vs. detected energy. 
            If scalar (float), efficiency is independent of energy. Value of `1.` (default) means 100% efficiency.
            If it's a vector (np.ndarray), the values correspond to the efficiency in each bin of `smearing.bins_smeared`.
            If it's a callable it should define efficiency as a function of energy
        weight:float.
            Channel weight, to be multiplied to the resulting event rates.
            Default value is 1.
        """
        self.flavor=flavor
        self.xsec = xsec
        self.efficiency = efficiency
        self.smearing = smearing
        self.weight = weight

    @property 
    def flavor(self):
        return self._flavor
    @flavor.setter
    def flavor(self, flavors):
        if isinstance(flavors, Flavor):
            flavors = [flavors]
        self._flavor = flavors
    @property
    def efficiency(self):
        return self._efficiency
        
    @efficiency.setter
    def efficiency(self, efficiency):
        if callable(efficiency) and not isinstance(efficiency,FunctionOfEnergy):
            efficiency = FunctionOfEnergy(efficiency)
        self._efficiency = efficiency

    @property
    def xsec(self):
        return self._xsec
    @xsec.setter
    def xsec(self, xsec):
        if callable(xsec) and not isinstance(xsec,FunctionOfEnergy):
                xsec = FunctionOfEnergy(xsec)
        self._xsec = xsec
    
    def __repr__(self):
        return f'{self.__class__.__name__} (flavor={",".join([f.name for f in self.flavor])}, smearing={self.smearing is not None}, weight={self.weight})'
    def calc_rate(self, flux:Container, apply_smearing=True, apply_efficiency=True)->Container:
        """Calculate the event rate in this channel
        
        Parameters
        ----------
        flux:Container
            Input neutrino flux
        apply_smearing:bool
            If `True`(default) apply the energy smearing, otherwise the result will be just interaction rate vs. neutrino energy
        apply_efficiency:bool
            If `True`(default) apply the efficiency, otherwise assume 100% efficiency
        """
        rate = self._calc_interaction_rate(flux)
        if apply_smearing:
            if self.smearing is not None:
                rate = self.smearing.apply(rate)
        if apply_efficiency:
                rate = rate*self.efficiency
        return rate
        
    def _calc_interaction_rate(self, flux):
        """calculate interaction rate for given channel"""
        tgt_mass = 1<<u.kt
        Ntargets = tgt_mass.to_value(u.Dalton)
        #sum flux over flavors
        array_total = sum([flux[flv].array for flv in self.flavor])
        #create a summary flux container
        flux_total = Container(array_total, flavor=self.flavor, 
                               time=flux.time, energy=flux.energy,
                               integrable_axes=flux._integrable_axes)
        rate = self.xsec*flux_total*self.weight*Ntargets
        return rate

class Detector:
    """A detector configuration for the rate calculation. """
    
    def __init__(self, name:str, mass: u.Quantity, channels: Dict[str,DetectionChannel]):
        """
        Parameters
        ----------
        name: str
            Detector name
        mass: Quantity[mass]
            Detector mass
        channels: Dict[str,DetectionChannel]
            Dictionary of detection channels in the format {name:channel}
    
        Note
        ----
        These parameters and detection channels can be modified later, before calling :meth:`run`
        """
        self.name = name
        self.mass = mass
        self.channels = channels

    def __repr__(self):
        return f'Detector(name="{self.name}", mass={self.mass}, channels={list(self.channels)})'
    
    def run(self, flux:Container, detector_effects:bool=True)->Dict[str, Container]:
        """Calculate the interaction rates for all channels in the detector.

        Parameters
        ----------
        flux:Container
            The incoming neutrino flux (or fluence).
        detector_effects:bool
            If `True` (default) apply the smearing and efficiency for all channels. Otherwise just calculate the interaction rates vs neutrino energy.

        Returns
        -------
        Dict[str,Container]
            Event rate for each detection channel in as dictionary {channel name: event rate}
        """
        result = {}
        for name,channel in self.channels.items():
            rate = channel.calc_rate(flux, apply_efficiency=detector_effects, apply_smearing=detector_effects)
            result[name] = rate*(self.mass/(1<<u.kt))
        return result

def _get_flavor_index(channel):
    _map = {'+e':Flavor.NU_E,
            '-e':Flavor.NU_E_BAR,
            '+m':Flavor.NU_X,
            '-m':Flavor.NU_X_BAR,
            '+t':Flavor.NU_X,
            '-t':Flavor.NU_X_BAR
            }
    return _map[channel.parity+channel.flavor]

def _bin_edges_from_centers(centers:np.ndarray)->np.ndarray:
    """calculate the bin edges based on the given central values"""
    binw = np.diff(centers) #get the bin width
    edges = centers-0.5*np.pad(binw,(0,1),mode='edge') #get lower edges
    edges = np.append(edges,edges[-1]+binw[-1])
    return edges
#--------------------------------------
class RateCalculator(SnowglobesData):
    r"""Simple rate calculation interface.
        Computes expected rate for a detector using SNOwGLoBES data. 
       
        Use :meth:`RateCalculator.run` method to run the simulation for specific detector and flux object.

        Input flux can either be energy differential flux :math:`F_\nu(E_i) = \frac{d N}{d_E}(E_i)` or integral flux in the energy bins :math:`N_{\nu}(E_i)`.
        
        If the `detector_effects=False` the result is the number of neutrino interactions :math:`N` - a product with with cross-section :math:`\sigma(E)` and the number of targets in the detector :math:`N_{tgt}`:

        .. math:: N (E_i) = N_{tgt} \cdot F_\nu(E_i)\cdot \sigma(E_i)


        If the detector effects are included, it is integrated within the energy bins and multiplied by the smearing matrix and detection efficiency to get number of interactions in energy bin:

        .. math:: 
            
            N_i = N_{tgt} \cdot \int\limits_{E_i}^{E_{i+1}}~F_\nu(E_i)~\sigma(E_i)~d E

        and this is convoluted with the detector smearing matrix :math:`M_{ij}` and efficiency :math:`\varepsilon(E)`:

        .. math::
        
            N^{rec}_i = \sum\limits_{j} M_{ij} \cdot N_j \cdot \varepsilon(E_i)
        
    """
    def __init__(self, base_dir=''):
        """
        Parameters
        ----------
        base_dir:         Path or None
            Path to the directory where the cross-section, detector, and channel files are located
            If empty, try to get it from ``$SNOWGLOBES`` environment var
        """
        super().__init__(base_dir=base_dir)

    def load_xsec(self, channel_name:str, flavor:Flavor)->FunctionOfEnergy:
        """Load cross-section for a given channel, interpolated in the energies"""
        xsec = np.loadtxt(self.base_dir/f"xscns/xs_{channel_name}.dat")
        # Cross-section in 10^-38 cm^2
        xp = xsec[:,0]
        #get the column to read from the file
        column = {Flavor.NU_E:1, Flavor.NU_X:2, Flavor.NU_E_BAR:4, Flavor.NU_X_BAR:5}[flavor]
        yp = xsec[:, column]
        def xsec(energies):
            E = energies.to_value('GeV')
            return np.interp(np.log(E)/np.log(10), xp, yp, left=0, right=0)*E*1e-38 <<u.cm**2
        return FunctionOfEnergy(xsec)
        
    def read_detector(self, name:str, material:str=None)->Detector:
        """Read the detector configuration from the SNOwGLoBES

        Parameters
        ----------
        name:str
            Detector name (see :attr:`detectors` for options)
        material:str or None
            Detector material (see :attr:`materials` for options)
            If `None` (default) try to guess material from detector name

        Returns
        -------
        Detector
            an object with the detector configuration.
        """
        material = material or guess_material(name)
        channels = {}
        bins = self.binning[material]
        bins_t = _bin_edges_from_centers(bins['e_true'])<<u.GeV
        bins_s = _bin_edges_from_centers(bins['e_smear'])<<u.GeV
        for ch in self.channels[material].itertuples():
            try:
                smearing = SmearingMatrix(bins_true=bins_t, 
                                          bins_smeared=bins_s,
                                          matrix=self.smearings[name][ch.name].T)
            except KeyError:
                warn(f'Smearing not found for detector={name}, channel={ch.name}. Using unsmeared spectrum')
                smearing = None
            try:            
                efficiency= self.efficiencies[name][ch.name]
            except KeyError:
                warn(f'Efficiency not found for detector={name}, channel={ch.name}. Using 100% efficiency')
                efficiency = 1
            flavor=_get_flavor_index(ch)
            channel = DetectionChannel(
                    flavor=flavor,
                    weight=ch.weight*self.detectors[name].factor,
                    xsec=self.load_xsec(ch.name,flavor),
                    smearing=smearing,
                    efficiency=efficiency)
            channels[ch.name]=channel
        
        return Detector(name=name,
                        mass=self.detectors[name].mass<<u.kt,
                        channels=channels
                       )
    def run(self, flux:Container, detector:str, material:str=None, detector_effects:bool = True)->Dict[str, Container]:
        """Run the rate calculation for the given detector.    
        
        Parameters
        ----------
        flux: Container
            The incoming neutrino flux (or fluence).
        
        detector: str
            Name of the detector to calculate the rate. 
            Check `RateCalculator.detectors` for the list of options
        
        material: str or None
            Name of the detector material. 
            Check `RateCalculator.materials` for the list of options
            If `None` (default) it will be guessed based on the name of the detector

        detector_effects: bool
            If true (default), account for efficiency and smearing. If false, consider a perfect detector.

        Returns
        -------
            dict[str, Container]
                A dictionary with interaction rates (as instances of :class:`snewpy.flux.Container`) for each channel.
        """
        return self.read_detector(detector,material).run(flux, detector_effects=detector_effects)


# -*- coding: utf-8 -*-
"""This module implements basic neutrino properties that are used throughout SNEWPY."""

from enum import IntEnum
from astropy import units as u
from dataclasses import dataclass
from typing import Optional
import numpy as np
from collections.abc import Mapping

class MassHierarchy(IntEnum):
    """Neutrino mass ordering: ``NORMAL`` or ``INVERTED``."""
    NORMAL = 1
    INVERTED = 2
    
    @classmethod
    def derive_from_dm2(cls, dm12_2, dm32_2, dm31_2):
        """derive the mass hierarchy based on the given mass squared differences"""
        assert dm12_2>0,f'dm12_2(dm12_2) should be positive'
        assert (dm32_2*dm31_2>=0),f'dm32_2 ({dm32_2}) and dm31_2 ({dm31_2}) should be of the same sign'
        if(dm32_2>=0):
            return MassHierarchy.NORMAL
        else:
            return MassHierarchy.INVERTED
            
class Flavor(IntEnum):
    """Enumeration of CCSN Neutrino flavors."""
    NU_E = 0
    NU_X = 1
    NU_E_BAR = 2
    NU_X_BAR = 3
    
    def to_tex(self):
        """LaTeX-compatible string representations of flavor."""
        if '_BAR' in self.name:
            return r'$\overline{{\nu}}_{0}$'.format(self.name[3].lower())
        return r'$\{0}$'.format(self.name.lower())

    @property
    def is_electron(self):
        """Return ``True`` for ``Flavor.NU_E`` and ``Flavor.NU_E_BAR``."""
        return self.value in (Flavor.NU_E.value, Flavor.NU_E_BAR.value)

    @property
    def is_neutrino(self):
        """Return ``True`` for ``Flavor.NU_E`` and ``Flavor.NU_X``."""
        return self.value in (Flavor.NU_E.value, Flavor.NU_X.value)

    @property
    def is_antineutrino(self):
        return self.value in (Flavor.NU_E_BAR.value, Flavor.NU_X_BAR.value)
        
@dataclass
class MixingParameters3Flavor(Mapping):
    """Mixing angles and mass differences, assuming three neutrino flavors.
    This class contains the default values used throughout SNEWPY, which are
    based on `NuFIT 5.0 <http://www.nu-fit.org>`_ results from July 2020,
    published in `JHEP 09 (2020) 178 <https://dx.doi.org/10.1007/JHEP09(2020)178>`_
    [`arXiv:2007.14792 <https://arxiv.org/abs/2007.14792>`_].
    """
    #mixing angles
    theta12: u.Quantity[u.deg]
    theta13: u.Quantity[u.deg]
    theta23: u.Quantity[u.deg]
    #CP violation phase         
    deltaCP: u.Quantity[u.deg]
    #square mass difference
    dm21_2: u.Quantity[u.eV**2]
    dm32_2: Optional[u.Quantity] = None
    dm31_2: Optional[u.Quantity] = None
    #mass ordering
    mass_order: Optional[MassHierarchy] = None
    # Note: in IH, the mass splittings are: m3..............m1..m2.

    def __iter__(self):
        return iter(self.__dict__)
    def __getitem__(self, item):
        return self.__dict__[item]
    def __len__(self):
        return len(self.__dict__)
    def _repr_markdown_(self):
        s = [f'**{self.__class__.__name__}**']
        s+=  ['|Parameter|Value|',
              '|:--------|:----:|']
        for name, v in self.__dict__.items():
            try: 
                s += [f"|{name}|{v._repr_latex_()}"]
            except:
                try: 
                    s += [f"|{name}|{v.name}"]
                except:
                    s += [f"|{name}|{v}|"]
        return '\n'.join(s)
        
    def __post_init__(self):
        #calculate the missing dm2
        if self.dm31_2 is None: 
            self.dm31_2 = self.dm32_2+self.dm21_2
        if self.dm32_2 is None: 
            self.dm32_2 = self.dm31_2-self.dm21_2  
        #evaluate mass ordering
        if self.mass_order is None:
            self.mass_order = MassHierarchy.derive_from_dm2(*self.get_mass_squared_differences())
        #validate angles
        #angles_sum = sum(self.get_mixing_angles())
        #assert angles_sum==90<<u.deg, f'Mixing angles sum is {angles_sum}!=90 degrees'
        #check hierarchy
        if self.mass_order==MassHierarchy.NORMAL:
            assert self.dm32_2>0, 'dm32_2 should be positive for NH'
            assert self.dm31_2>0, 'dm31_2 should be positive for NH'
        else:
            assert self.dm32_2<0, 'dm32_2 should be negative for IH'
            assert self.dm31_2<0, 'dm31_2 should be negative for IH'
        #validate dm2
        dm2_sum = self.dm32_2+self.dm21_2-self.dm31_2
        assert np.isclose(dm2_sum,0), f'dm32_2+dm31_2-dm31_2 = {dm2_sum} !=0'

    def get_mixing_angles(self):
        """Mixing angles of the PMNS matrix.
        
        Returns
        -------
        tuple
            Angles theta12, theta13, theta23.
        """
        return (self.theta12, self.theta13, self.theta23)
        
    def get_mass_squared_differences(self):
        """Mass squared differences .
        
        Returns
        -------
        tuple
            dm21_2, dm31_2, dm32_2.
        """        
        return (self.dm21_2, self.dm31_2, self.dm32_2)

@dataclass
class MixingParameters4Flavor(MixingParameters3Flavor):
    """A class for four flavor neutrino mixing. 
    ..Note: it is an extension of :class:`MixingParameters3Flavor`, and can be constructed using it:
    
        >>> pars_3f = MixingParameters() #standard 3flavor mixing
        >>> pars_4f = MixingParameters4Flavor(**pars_3f, theta14=90<<u.deg, dm41_2=1<<u.GeV**2)
    """
    #sterile neutrino miging angles
    theta14: u.Quantity[u.deg] = 0<<u.deg
    theta24: u.Quantity[u.deg] = 0<<u.deg
    theta34: u.Quantity[u.deg] = 0<<u.deg
    #sterile neutrino mass squared differences
    dm41_2: u.Quantity[u.eV**2] = 0<<u.eV**2
    dm42_2: Optional[u.Quantity] = None
    dm43_2: Optional[u.Quantity] = None
    
    def __post_init__(self):
        super().__post_init__()
        self.dm42_2 = self.dm42_2 or self.dm41_2 - self.dm21_2
        self.dm43_2 = self.dm43_2 or self.dm41_2 - self.dm31_2

        dm2_sum = self.dm41_2 - self.dm42_2 - self.dm21_2
        assert np.isclose(dm2_sum,0), f'dm41_2 - dm42_2 - dm21_2 = {dm2_sum} !=0'
        dm2_sum = self.dm41_2 - self.dm43_2 - self.dm31_2
        assert np.isclose(dm2_sum,0), f'dm41_2 - dm43_2 - dm31_2 = {dm2_sum} !=0'

parameter_presets = {
    'NuFIT5.0': {
        # Values from http://www.nu-fit.org/?q=node/228; cite as JHEP 09 (2020) 178 [arXiv:2007.14792]
        MassHierarchy.NORMAL:
        MixingParameters3Flavor(
            theta12 = 33.44 << u.deg,
            theta13 = 8.57 << u.deg,
            theta23 = 49.2 << u.deg,
            deltaCP = 197 << u.deg,
            dm21_2 = 7.42e-5 << u.eV**2,
            dm31_2 = 2.517e-3 << u.eV**2
        ),
        MassHierarchy.INVERTED:
        MixingParameters3Flavor(
            theta12 = 33.45 << u.deg,
            theta13 = 8.60 << u.deg,
            theta23 = 49.3 << u.deg,
            deltaCP = 282 << u.deg,
            dm21_2 = 7.42e-5 << u.eV**2,
            dm32_2 = -2.498e-3 << u.eV**2
        )
    },
    'NuFIT5.2': {
        # Values from http://www.nu-fit.org/?q=node/256; cite as JHEP 09 (2020) 178 [arXiv:2007.14792]
        MassHierarchy.NORMAL:
        MixingParameters3Flavor(
            theta12 = 33.41 << u.deg,
            theta13 = 8.58 << u.deg,
            theta23 = 42.2 << u.deg,
            deltaCP = 232 << u.deg,
            dm21_2 = 7.41e-5 << u.eV**2,
            dm31_2 = 2.507e-3 << u.eV**2
        ),
        MassHierarchy.INVERTED:
        MixingParameters3Flavor(
            theta12 = 33.41 << u.deg,
            theta13 = 8.57 << u.deg,
            theta23 = 49.0 << u.deg,
            deltaCP = 276 << u.deg,
            dm21_2 = 7.41e-5 << u.eV**2,
            dm32_2 = -2.486e-3 << u.eV**2
        )
    },
    'NuFIT6.0': {
        # Values from http://www.nu-fit.org/?q=node/294; cite as arXiv:2410.05380
        MassHierarchy.NORMAL:
        MixingParameters3Flavor(
            theta12 = 33.68 << u.deg,
            theta13 = 8.56 << u.deg,
            theta23 = 43.3 << u.deg,
            deltaCP = 212 << u.deg,
            dm21_2 = 7.49e-5 << u.eV**2,
            dm31_2 = 2.513e-3 << u.eV**2
        ),
        MassHierarchy.INVERTED:
        MixingParameters3Flavor(
            theta12 = 33.68 << u.deg,
            theta13 = 8.59 << u.deg,
            theta23 = 47.9 << u.deg,
            deltaCP = 274 << u.deg,
            dm21_2 = 7.49e-5 << u.eV**2,
            dm32_2 = -2.484e-3 << u.eV**2
        )
    },
    'PDG2022':{
        # Cite as R.L. Workman et al. (Particle Data Group), Prog. Theor. Exp. Phys. 2022, 083C01 (2022)
        MassHierarchy.NORMAL:
        MixingParameters3Flavor(
            theta12 = 33.65 << u.deg,
            theta13 = 8.53 << u.deg,
            theta23 = 47.64 << u.deg,
            deltaCP = 245 << u.deg,
            dm21_2 = 7.53e-5 << u.eV**2,
            dm32_2 = 2.453e-3 << u.eV**2
        ),
        MassHierarchy.INVERTED:
        MixingParameters3Flavor(
            theta12 = 33.65 << u.deg,
            theta13 = 8.53 << u.deg,
            theta23 = 47.24 << u.deg,
            deltaCP = 245 << u.deg,
            dm21_2 = 7.53e-5 << u.eV**2,
            dm32_2 = -2.536e-3 << u.eV**2
        )
    },
    'PDG2024':{
        # Values from https://pdglive.lbl.gov/Particle.action?node=S067&init=0
        # Cite as S. Navas et al. (Particle Data Group), Phys. Rev. D 110, 030001 (2024)
        MassHierarchy.NORMAL:
        MixingParameters3Flavor(
            theta12 = 33.65 << u.deg,
            theta13 = 8.51 << u.deg,
            theta23 = 48.33 << u.deg,
            deltaCP = 214 << u.deg,
            dm21_2 = 7.53e-5 << u.eV**2,
            dm32_2 = 2.455e-3 << u.eV**2
        ),
        MassHierarchy.INVERTED:
        MixingParameters3Flavor(
            theta12 = 33.65 << u.deg,
            theta13 = 8.51 << u.deg,
            theta23 = 48.04 << u.deg,
            deltaCP = 214 << u.deg,
            dm21_2 = 7.53e-5 << u.eV**2,
            dm32_2 = -2.529e-3 << u.eV**2
        )
    }
}
   

def MixingParameters(mh:MassHierarchy=MassHierarchy.NORMAL, version:str='NuFIT5.0'):
    return parameter_presets[version][mh]



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
======
snewpy
======
Front-end for supernova models which provide neutrino luminosity and spectra.
"""

from sys import exit
import os

try:
    from astropy.config.paths import get_cache_dir
except ModuleNotFoundError:
    # when imported by setup.py before dependencies are installed
    get_cache_dir = lambda: '.'


__version__ = '1.6'


src_path = os.path.realpath(__path__[0])
base_path = os.sep.join(src_path.split(os.sep)[:-2])
model_path = os.path.join(get_cache_dir(), 'snewpy', 'models')

def get_models(models=None, download_dir=None):
    """Download model files from the snewpy repository.

    Parameters
    ----------
    models : list or str
        Models to download. Can be 'all', name of a single model or list of model names.
    download_dir : str
        [Deprecated, do not use.]
    """
    from concurrent.futures import ThreadPoolExecutor, as_completed
    from warnings import warn
    from .models.registry_model import all_models

    if download_dir is not None:
        warn("The `download_dir` argument to `get_models` is deprecated and will be removed soon.", FutureWarning, stacklevel=2)

    all_models = {m.__name__: m for m in all_models}
    all_model_names = sorted(all_models.keys())

    if models == "all":
        models = all_model_names
    elif isinstance(models, str):
        models = [models]
    elif models is None:
        # Select model(s) to download
        print(f"Available models in SNEWPY v{__version__}: {all_model_names}")

        selected = input("\nType a model name, 'all' to download all models or <Enter> to cancel: ").strip()
        if selected == "all":
            models = all_model_names
        elif selected == "":
            exit()
        elif selected in all_model_names:
            models = {selected}
            while True:
                selected = input("\nType another model name or <Enter> if you have selected all models you want to download: ").strip()
                if selected in all_model_names:
                    models.add(selected)
                elif selected == "":
                    break
                else:
                    print(f"'{selected}' is not a valid model name. Please check for typos.")
        else:
            print(f"'{selected}' is not a valid model name. Please check for typos.")
            exit()

        print(f"\nYou have selected the models: {models}\n")

    pool = ThreadPoolExecutor(max_workers=8)
    results = []
    print(f"Downloading files for {models} to '{model_path}' ...")
    for model in models:
        for progenitor in all_models[model].get_param_combinations():
            results.append(pool.submit(all_models[model], **progenitor))

    exceptions = []
    for result in as_completed(results):
        if result.exception() is not None:
            exceptions.append(result.exception())
    if exceptions:
        print(f"ERROR: {len(exceptions)} exceptions occured. ({exceptions})")
        print("Please check your internet connection and try again later. If this persists, please report it at https://github.com/SNEWS2/snewpy/issues")
        exit(1)
    pool.shutdown(wait=False)
